@@ -46,134 +46,17 @@
 #include <dirent.h>
 #include <extractor.h>
 #include <locale.h>
-
-#include "gettext.h"
-#define _(String) gettext (String)
-#define gettext_noop(String) String
-
-#ifndef MINGW
- #define DIR_SEPARATOR '/'
- #define DIR_SEPARATOR_STR "/"
-#else
- #define DIR_SEPARATOR '\\'
- #define DIR_SEPARATOR_STR "\\"
-#endif
+#include <langinfo.h>
 
 #include "doodle.h"
 #include "getopt.h"
-#include "grow.c"
+#include "helper2.h"
+#include "convert.h"
 
-/**
- * Maximum search-string length.  If a search-string is more than
- * MAX_LENGTH/2 characters long, it may not find a match even if one
- * exists.  This value can be enlarged arbitrarily, but the cost is
- * that building the database will become more expensive IF for
- * keywords that are that long (the cost for search-strings is
- * quadratic in the length of the search-string; with MAX_LENGTH it is
- * bounded to become linear after a certain size by breaking the
- * string into smaller sections).
- */
-#define MAX_LENGTH 128
-
-typedef struct {
-  char shortArg;
-  char * longArg;
-  char * mandatoryArg;
-  char * description;
-} Help;
-
-#define BORDER 29
 
 static int verbose = 0;
 static int very_verbose = 0;
 
-/**
- * Produce nicely formatted output for --help.  Generic
- * method also used in libextractor and GNUnet.
- */
-static void formatHelp(const char * general,
-		       const char * description,
-		       const Help * opt) {
-  int slen;
-  int i;
-  int j;
-  int ml;
-  int p;
-  char * scp;
-  const char * trans;
-	   
-  printf(_("Usage: %s\n%s\n\n"),
-	 gettext(general),
-	 gettext(description));
-  printf(_("Arguments mandatory for long options are also mandatory for short options.\n"));
-  slen = 0;
-  i = 0;
-  while (opt[i].description != NULL) {
-    if (opt[i].shortArg == 0)
-      printf("      ");
-    else
-      printf("  -%c, ",
-	     opt[i].shortArg);
-    printf("--%s",
-	   opt[i].longArg);
-    slen = 8 + strlen(opt[i].longArg);
-    if (opt[i].mandatoryArg != NULL) {
-      printf("=%s",
-	     opt[i].mandatoryArg);
-      slen += 1+strlen(opt[i].mandatoryArg);
-    }    
-    if (slen > BORDER) {
-      printf("\n%*s", BORDER, "");
-      slen = BORDER;
-    }
-    if (slen < BORDER) {
-      printf("%*s", BORDER-slen, "");
-      slen = BORDER;
-    }
-    trans = gettext(opt[i].description);
-    ml = strlen(trans);
-    p = 0;
-  OUTER:
-    while (ml - p > 78 - slen) {
-      for (j=p+78-slen;j>p;j--) {
-	if (isspace(trans[j])) {
-	  scp = MALLOC(j-p+1);
-	  memcpy(scp,
-		 &trans[p],
-		 j-p);
-	  scp[j-p] = '\0';
-	  printf("%s\n%*s",
-		 scp,
-		 BORDER+2,
-		 "");
-	  free(scp);
-	  p = j+1;
-	  slen = BORDER+2;
-	  goto OUTER;
-	}
-      }
-      /* could not find space to break line */
-      scp = MALLOC(78 - slen + 1);
-      memcpy(scp,
-	     &trans[p],
-	     78 - slen);
-      scp[78 - slen] = '\0';
-      printf("%s\n%*s",
-	     scp,
-	     BORDER+2,
-	     "");	
-      free(scp);
-      slen = BORDER+2;
-      p = p + 78 - slen;
-    }
-    /* print rest */
-    if (p < ml)
-      printf("%s\n",
-	     &trans[p]);
-    i++;
-  }
-}
-
 /**
  * Print the doodle-specific text for --help.
  */
@@ -228,7 +111,24 @@ static int ignore_case = 0;
 static unsigned int do_approx = 0;
 static char * prunepaths = "/tmp /usr/tmp /var/tmp /dev /proc";
 
-/* *************** directory traversals **************** */
+/* *************** helper functions **************** */
+
+/**
+ * Print log-messages to stdout.
+ */
+static void my_log(void * unused,
+		   unsigned int level,
+		   const char * msg,
+		   ...) {
+  va_list args;
+  if ( (level == 0) ||
+       (verbose && (level == 1)) ||
+       (very_verbose && (level == 2) ) ) {    
+    va_start(args, msg);
+    vfprintf(stdout, msg, args);
+    va_end(args);
+  }    
+}
 
 static int isPruned(const char * filename) {
   int i;
@@ -252,303 +152,9 @@ static int isPruned(const char * filename) {
   return 0;
 }
 
-typedef int (*ScannerCallback)(const char * filename,
-			       void * arg);
-
-/**
- * Scan a directory for files. The name of the directory
- * must be expanded first (!).
- * @param dirName the name of the directory
- * @param callback the method to call for each file,
- *        can be NULL, in that case, we only count
- * @param data argument to pass to callback
- * @return the number of files found, -1 on error
- */
-static int scanDirectory(char * dirName,
-			 int verbose,
-			 int very_verbose,
-			 ScannerCallback callback,
-			 void * cb_arg) {
-  DIR * dinfo;
-  struct dirent *finfo;
-  struct stat istat;
-  int count = 0;
-
-  if (dirName == NULL) {
-    /* dirName==NULL */
-    printf(_("Assertion failed at %s:%d.\n"),
-	   __FILE__, __LINE__);
-    return -1;
-  }
-  if (isPruned(dirName))
-    return 0;
-  if (0 != lstat(dirName, &istat)) {
-    if (verbose) 
-      printf(_("Call to '%s' for file '%s' failed: %s\n"),
-	     "lstat",
-	     dirName, 
-	     strerror(errno));
-    return 0;
-  }
-#ifdef S_ISLNK
-  if (S_ISLNK(istat.st_mode))
-    return 0;
-#endif
-#ifdef S_ISSOCK
-  if (S_ISSOCK(istat.st_mode))
-    return 0;
-#endif
-  if (S_ISCHR(istat.st_mode) || 
-      S_ISBLK(istat.st_mode) ||
-      S_ISFIFO(istat.st_mode) )
-    return 0;
-  if (S_ISREG(istat.st_mode)) {
-    if (-1 == callback(dirName,
-		       cb_arg))
-      return -1;
-    else
-      return 1;
-  }
-  if (! S_ISDIR(istat.st_mode))
-    return 0;
-  if (verbose)
-    printf(_("Scanning '%s'\n"), dirName);
-  errno = 0;
-  dinfo = opendir(dirName);
-  if ((errno == EACCES) || (dinfo == NULL)) {
-    if ( (verbose) &&
-	 (errno == EACCES) )
-      printf(_("Access to directory '%s' was denied.\n"),
-	     dirName);
-    return 0;
-  }
-  while ((finfo = readdir(dinfo)) != NULL) {
-    char * dn;
-    int ret;
-
-    if (finfo->d_name[0] == '.')
-      continue;
-    dn = MALLOC(strlen(dirName) + 
-		strlen(finfo->d_name) + 2);
-    strcpy(dn, dirName);
-    if (dirName[strlen(dirName)-1] != DIR_SEPARATOR)
-      strcat(dn, DIR_SEPARATOR_STR);
-    strcat(dn, finfo->d_name);
-    ret = scanDirectory(dn,
-			verbose,
-			very_verbose,
-			callback,
-			cb_arg);   
-    free(dn);
-    if (ret >= 0) 
-      count += ret;
-    else
-      return -1;
-  }
-  closedir(dinfo);
-  return count;
-}
-
-
-/**
- * Complete filename (a la shell) from abbrevition.
- * @param fil the name of the file, may contain ~/ or 
- *        be relative to the current directory
- * @returns the full file name, 
- *          NULL is returned on error
- */
-static char * expandFileName(const char * fil) {
-  char buffer[512];
-  char * fn;
-#ifndef MINGW
-  char * fm;
-  const char *fil_ptr;
-#else
-  long lRet;
-#endif
-
-  if (fil == NULL)
-    return NULL;
-
-#ifndef MINGW
-  if (fil[0] == DIR_SEPARATOR) {
-    /* absolute path, just copy */
-    return STRDUP(fil);
-  }
-  if (fil[0] == '~') {
-    fm = getenv("HOME");
-    if (fm == NULL) {
-      /* keep it symbolic to show error to user! */
-      fm = "$HOME";
-    }
-
-    /* do not copy '~' */
-    fil_ptr = fil + 1;
-
-	/* skip over dir seperator to be consistent */    
-    if (fil_ptr[0] == DIR_SEPARATOR) 
-      fil_ptr++;    
-  } else {
-    fil_ptr = fil;
-    if (getcwd(buffer, 512) != NULL)
-      fm = buffer;
-    else
-      fm = "$PWD";
-  }
-  fn = MALLOC(strlen(fm) + 1 + strlen(fil_ptr) + 1);
-  
-  sprintf(fn, "%s/%s", fm, fil_ptr);
-#else
-  fn = MALLOC(MAX_PATH + 1);
-
-  if ((lRet = conv_to_win_path(fil, buffer)) != ERROR_SUCCESS)
-  {
-  	SetErrnoFromWinError(lRet);
-
-    return NULL;
-  }
-  /* is the path relative? */
-  if ((strncmp(buffer + 1, ":\\", 2) != 0) && 
-      (strncmp(buffer, "\\\\", 2) != 0))
-  {
-    char szCurDir[MAX_PATH + 1];
-    lRet = GetCurrentDirectory(MAX_PATH + 1, szCurDir);
-    if (lRet + strlen(fn) + 1 > (_MAX_PATH + 1))
-    {
-      SetErrnoFromWinError(ERROR_BUFFER_OVERFLOW);
-      
-      return NULL;
-    }
-    sprintf(fn, "%s\\%s", szCurDir, buffer);
-  }
-  else
-  {
-    strcpy(fn, buffer);
-  }
-#endif
-  return fn;
-}
-
-/**
- * Scan location for files.  Call callback on each file.  Abort if
- * callback returns -1.
- *
- * @return -1 on error
- */
-static int scan(char * location,
-		ScannerCallback callback,
-		void * cb_arg) {
-  int ret;
-  char * exp;
-  
-  exp = expandFileName(location);
-  ret = scanDirectory(exp,
-		      verbose,
-		      very_verbose,
-		      callback,
-		      cb_arg);
-  free(exp);
-  return ret;
-}
-
-
 
 /* ************** actual functionality code ************* */
 
-/**
- * Find keywords in the given file and append
- * the string describing these keywords to the
- * long string (for the suffix tree).
- */
-static int buildIndex(EXTRACTOR_ExtractorList * elist,
-		      const char * filename,
-		      FILE * logFile,
-		      struct DOODLE_SuffixTree * tree) {
-  EXTRACTOR_KeywordList * head;
-  EXTRACTOR_KeywordList * pos;
-
-  head = EXTRACTOR_getKeywords(elist,
-			       filename);
-  head = EXTRACTOR_removeDuplicateKeywords
-    (head,
-     EXTRACTOR_DUPLICATES_REMOVE_UNKNOWN);
-  pos  = head;
-  if (do_extract) {
-    printf(_("Keywords for '%s':\n"),
-	   filename);
-    EXTRACTOR_printKeywords(stdout,
-			    head);
-  }
-  while (pos != NULL) {
-    char * cpos = pos->keyword;
-    size_t slen;
-
-    if (logFile != NULL)
-      fprintf(logFile, "%s\n", cpos);
-    slen = strlen(cpos);
-    if (slen > MAX_LENGTH) {
-      char section[MAX_LENGTH+1];
-      char * xpos;
-      int j;
-
-      for (j=0;j<slen;j+=MAX_LENGTH/2) {
-	strncpy(section,
-		&cpos[j],
-		MAX_LENGTH);
-	xpos = &section[0];
-	while (xpos[0] != '\0') {
-	  if (0 != DOODLE_tree_expand(tree,
-				      xpos,
-				      filename)) {
-	    EXTRACTOR_freeKeywords(head);
-	    return 0;
-	  }
-	  xpos++;
-	}
-      }
-    } else {
-      while (cpos[0] != '\0') {
-	if (0 != DOODLE_tree_expand(tree,
-				    cpos,
-				    filename)) {
-	  EXTRACTOR_freeKeywords(head);
-	  return 0;
-	}
-	cpos++;
-      }
-    }
-    pos = pos->next;
-  }
-  EXTRACTOR_freeKeywords(head);
-  if (do_filenames) {
-    const char * cpos = filename;
-    while (cpos[0] != '\0') {
-      if (0 != DOODLE_tree_expand(tree,
-				  cpos,
-				  filename)) 
-	return 0;
-      cpos++;
-    }
-  }    
-  return 1;
-}
-
-/**
- * Print log-messages to stdout.
- */
-static void my_log(void * unused,
-		   unsigned int level,
-		   const char * msg,
-		   ...) {
-  va_list args;
-  if ( (level == 0) ||
-       (verbose && (level == 1)) ||
-       (very_verbose && (level == 2) ) ) {    
-    va_start(args, msg);
-    vfprintf(stdout, msg, args);
-    va_end(args);
-  }    
-}
 
 /**
  * @brief closure for do_index
@@ -565,7 +171,9 @@ static int do_index(const char * filename,
   int i;
   int j;
   struct stat sbuf;
-  
+
+  if (isPruned(filename))
+    return 0; 
   j = -1;
   for (i=DOODLE_getFileCount(dic->tree)-1;i>=0;i--) 
     if (0 == strcmp(filename,
@@ -573,30 +181,26 @@ static int do_index(const char * filename,
       j = i;
       break;
     }
-  if (j != -1) {
-    if (0 != stat(filename,
-		  &sbuf)) {
-      printf(_("Call to '%s' for file '%s' failed: %s\n"),
-	     "stat",
-	     filename,
-	     strerror(errno));
-      return -1;
-    }
-    if (DOODLE_getFileAt(dic->tree,j)->mod_time == (unsigned int) sbuf.st_mtime) {
-      return 0; /* already processed! */
-    } else {
-      /* remove old keywords, file changed! */
-      DOODLE_tree_truncate(dic->tree,
-			   filename);
-    }
+  if (j != -1) 
+    return 0; /* already processed */
+  if (0 != stat(filename,
+		&sbuf)) {
+    printf(_("Call to '%s' for file '%s' failed: %s\n"),
+	   "stat",
+	   filename,
+	   strerror(errno));
+    return 0;
   }
-  if (very_verbose)
-    printf(_("Processing '%s'\n"),
-	   filename);
-  return buildIndex(dic->elist,
-		    filename,
-		    dic->logFile,
-		    dic->tree);
+  if (S_ISREG(sbuf.st_mode)) {
+    if (very_verbose)
+      printf(_("Processing '%s'\n"),
+	     filename);
+    return buildIndex(dic->elist,
+		      filename,
+		      dic->tree,
+		      do_filenames);
+  } else
+    return 0;
 }
 
 static int build(const char * libraries,
@@ -636,31 +240,9 @@ static int build(const char * libraries,
     DOODLE_tree_set_memory_limit(cls.tree,
 				 mem_limit);
 
-  /* first check if all indexed files still exist and are
-     accessible */
-  if (verbose)
-    printf(_("Checking integrity of existing database.\n"));
-  for (i=DOODLE_getFileCount(cls.tree)-1;i>=0;i--) {
-    struct stat sbuf;
-    if ( (0 != lstat(DOODLE_getFileAt(cls.tree,i)->filename,
-		     &sbuf)) &&
-	 ( (errno == ENOENT) ||
-	   (errno == ENOTDIR) ||
-	   (errno == ELOOP) ||
-	   (errno == EACCES) ) ) {
-      printf(_("File '%s' could not be accessed: %s. Removing file from index.\n"),
-	     DOODLE_getFileAt(cls.tree,i)->filename,
-	     strerror(errno));
-      DOODLE_tree_truncate(cls.tree,
-			   DOODLE_getFileAt(cls.tree,i)->filename);
-    }
-    if (! S_ISREG(sbuf.st_mode)) {
-      printf(_("File '%s' is not a regular file. Removing file from index.\n"),
-	     DOODLE_getFileAt(cls.tree,i)->filename);
-      DOODLE_tree_truncate(cls.tree,
-			   DOODLE_getFileAt(cls.tree,i)->filename);
-    }
-  }
+  DOODLE_tree_truncate_modified(cls.tree,
+			       &my_log,
+			       NULL);
 
   if (do_default) {
     if (verbose)
@@ -691,15 +273,22 @@ static int build(const char * libraries,
 
   ret = 0;
   for (i=0;i<argc;i++) {
+    char * exp;
+
     if (verbose)
       printf(_("Indexing '%s'\n"),
 	     argv[i]);
-    if (-1 == scan(argv[i],
-		   (ScannerCallback) &do_index,
-		   &cls)) {
+    exp = expandFileName(argv[i]);
+    if (-1 == scanDirectory(exp,
+			    &my_log,
+			    NULL,
+			    (ScannerCallback) &do_index,
+			    &cls)) {
       ret = -1;
+      free(exp);
       break;
     }
+    free(exp);   
   }
   EXTRACTOR_removeAll(cls.elist);
   DOODLE_tree_destroy(cls.tree);
@@ -788,9 +377,9 @@ static int print(const char * dbName) {
     free(ename);
     return -1;
   }
-  tree = DOODLE_tree_create(&my_log,
-			    NULL,
-			    ename);
+  tree = DOODLE_tree_open_RDONLY(&my_log,
+				 NULL,
+				 ename);
   free(ename);
   if (tree == NULL) 
     return -1;
@@ -812,6 +401,7 @@ static int search(const char * libraries,
   EXTRACTOR_ExtractorList * extractors;
   int i;
   PrintItArgs args;
+  char * utf;
 
   if (dbName == NULL) {
     printf(_("No database specified. Aborting.\n"));
@@ -826,9 +416,9 @@ static int search(const char * libraries,
     free(ename);
     return -1;
   }
-  tree = DOODLE_tree_create(&my_log,
-			    NULL,
-			    ename);
+  tree = DOODLE_tree_open_RDONLY(&my_log,
+				 NULL,
+				 ename);
   if (mem_limit != 0)
     DOODLE_tree_set_memory_limit(tree,
 				 mem_limit);
@@ -865,10 +455,13 @@ static int search(const char * libraries,
       printf(_("Warning: search string is longer than %d characters, search may not work properly.\n"),
 	     MAX_LENGTH/2);
     }
+    utf = convertToUtf8(argv[i],
+			strlen(argv[i]),
+			nl_langinfo(CODESET));
     if ( (do_approx == 0) &&
 	 (ignore_case == 0) ) {
       if (0 == DOODLE_tree_search(tree,
-				  argv[i],
+				  utf,
 				  (DOODLE_ResultCallback) &printIt,
 				  &args)) {
 	printf(_("\tNot found!\n"));
@@ -878,13 +471,14 @@ static int search(const char * libraries,
       if (0 == DOODLE_tree_search_approx(tree,
 					 do_approx,
 					 ignore_case,
-					 argv[i],
+					 utf,
 					 (DOODLE_ResultCallback) &printIt,
 					 &args)) {
 	printf(_("\tNot found!\n"));
 	ret++;
       }
     }
+    free(utf);
   }  
   DOODLE_tree_destroy(tree);
   EXTRACTOR_removeAll(extractors);
